/*
 * Copyright (c) 1999-2007 Apple Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 * 
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */

#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <arpa/nameser.h>
#include "dns_util.h"

/*
 * Free a resource record structure.
 */
void dns_free_resource_record(dns_resource_record_t *rr);

static uint8_t
_dns_parse_uint8(char **p)
{
	uint8_t v;

	v = (uint8_t)**p;
	*p += 1;
	return v;
}

static uint16_t
_dns_parse_uint16(char **p)
{
	uint16_t *x, v;

	x = (uint16_t *)*p;
	v = ntohs(*x);
	*p += 2;
	return v;
}

static uint32_t
_dns_parse_uint32(char **p)
{
	uint32_t *x, v;

	x = (uint32_t *)*p;
	v = ntohl(*x);
	*p += 4;
	return v;
}

static char *
_dns_parse_string(const char *p, char **x, int32_t *remaining)
{
	char *str;
	uint8_t len;

	if (*remaining < 1) return NULL;
	*remaining -= 1;

	len = (uint8_t)**x;
	*x += 1;

	if (*remaining < len) return NULL;
	*remaining -= len;

	str = malloc(len + 1);
	memmove(str, *x, len);
	str[len] = '\0';
	*x += len;

	return str;
}

static char *
_dns_parse_domain_name(const char *p, char **x, int32_t *remaining)
{
	uint8_t *v8;
	uint16_t *v16, skip;
	uint16_t i, j, dlen, len;
	int more, compressed;
	char *name, *start, *y, *z;

	if (*remaining < 1) return NULL;

	z = *x + *remaining;
	start = *x;
	compressed = 0;
	more = 1;
	name = malloc(1);
    if (name == NULL) {
        return NULL;
    }
	name[0] = '\0';
	len = 1;
	j = 0;
	skip = 0;

	while (more == 1)
	{
		if ((*x + 1) > z)
		{
			free(name);
			return NULL;
		}

		v8 = (uint8_t *)*x;
		dlen = *v8;

		if ((dlen & 0xc0) == 0xc0)
		{
			if ((*x + 2) > z)
			{
				free(name);
				return NULL;
			}

			v16 = (uint16_t *)*x;

			y = (char *)p + (ntohs(*v16) & 0x3fff);
			if ((*x <= y) || (y > z))
			{
				free(name);
				return NULL;
			}

			*x = y;
			if (compressed == 0) skip += 2;
			compressed = 1;
			continue;
		}

		if ((*x + 1) > z)
		{
			free(name);
			return NULL;
		}

		*x += 1;

		if (dlen > 0)
		{
            if (len > UINT16_MAX - dlen) {
                free(name);
                return NULL;
            }
			len += dlen;
			name = reallocf(name, len);
            if (name == NULL) {
                return NULL;
            }
		}

		if ((*x + dlen) > z)
		{
			free(name);
			return NULL;
		}

        if (len > UINT16_MAX - dlen - 1) {
            free(name);
            return NULL;
        }
        
		for (i = 0; i < dlen; i++)
		{
			name[j++] = **x;
			*x += 1;
		}
		name[j] = '\0';
        
		if (compressed == 0) skip += (dlen + 1);

		if (dlen == 0) more = 0;
		else
		{
			if ((*x + 1) > z)
			{
				free(name);
				return NULL;
			}

			v8 = (uint8_t *)*x;
			if (*v8 != 0)
			{
                if ((len < 1) || (len > UINT16_MAX - 2)) {
                    free(name);
                    return NULL;
                }
                
				len += 1;
				name = reallocf(name, len);
                if (name == NULL) {
                    return NULL;
                }
				name[j++] = '.';
				name[j] = '\0';
			}
		}
	}

	if ((start + skip) > z)
	{
		free(name);
		return NULL;
	}

	*x = start + skip;
	*remaining -= skip;

	return name;
}

static
dns_resource_record_t *
_dns_parse_resource_record_internal(const char *p, char **x, int32_t *remaining)
{
	uint32_t size, bx, mi;
	uint16_t rdlen;
	uint8_t byte, i;
	dns_resource_record_t *r;
	char *eor;
    int32_t rem;

	if (*remaining < 1) return NULL;

	r = (dns_resource_record_t *)calloc(1, sizeof(dns_resource_record_t));

	r->name = _dns_parse_domain_name(p, x, remaining);
	if (r->name == NULL)
	{
		dns_free_resource_record(r);
		return NULL;
	}

	if (*remaining < 10)
	{
		dns_free_resource_record(r);
		return NULL;
	}

	r->dnstype = _dns_parse_uint16(x);
	r->dnsclass = _dns_parse_uint16(x);
	r->ttl = _dns_parse_uint32(x);
	rdlen = _dns_parse_uint16(x);

	*remaining -= 10;

	if (*remaining < rdlen)
	{
		dns_free_resource_record(r);
		return NULL;
	}

	eor = *x;
    rem = *remaining;
	r->data.A = NULL;

	switch (r->dnstype)
	{
		case ns_t_a:
			if (*remaining < 4)
			{
				dns_free_resource_record(r);
				return NULL;
			}

			*remaining -= 4;

			size = sizeof(dns_address_record_t);
			r->data.A = (dns_address_record_t *)calloc(1, size);
			r->data.A->addr.s_addr = htonl(_dns_parse_uint32(x));
			break;

		case ns_t_aaaa:
			if (*remaining < 16)
			{
				dns_free_resource_record(r);
				return NULL;
			}

			*remaining -= 16;

			size = sizeof(dns_in6_address_record_t);
			r->data.AAAA = (dns_in6_address_record_t *)calloc(1, size);
			r->data.AAAA->addr.__u6_addr.__u6_addr32[0] = htonl(_dns_parse_uint32(x));
			r->data.AAAA->addr.__u6_addr.__u6_addr32[1] = htonl(_dns_parse_uint32(x));
			r->data.AAAA->addr.__u6_addr.__u6_addr32[2] = htonl(_dns_parse_uint32(x));
			r->data.AAAA->addr.__u6_addr.__u6_addr32[3] = htonl(_dns_parse_uint32(x));
			break;

		case ns_t_ns:
		case ns_t_md:
		case ns_t_mf:
		case ns_t_cname:
		case ns_t_mb:
		case ns_t_mg:
		case ns_t_mr:
		case ns_t_ptr:
			size = sizeof(dns_domain_name_record_t);
			r->data.CNAME = (dns_domain_name_record_t *)calloc(1,size);
			r->data.CNAME->name = _dns_parse_domain_name(p, x, remaining);
			if (r->data.CNAME->name == NULL)
			{
				dns_free_resource_record(r);
				return NULL;
			}
			break;

		case ns_t_soa:
			size = sizeof(dns_SOA_record_t);
			r->data.SOA = (dns_SOA_record_t *)calloc(1, size);

			r->data.SOA->mname = _dns_parse_domain_name(p, x, remaining);
			if (r->data.SOA->mname == NULL)
			{
				dns_free_resource_record(r);
				return NULL;
			}

			r->data.SOA->rname = _dns_parse_domain_name(p, x, remaining);
			if (r->data.SOA->rname == NULL)
			{
				dns_free_resource_record(r);
				return NULL;
			}

			if (*remaining < 20)
			{
				dns_free_resource_record(r);
				return NULL;
			}

			*remaining -= 20;

			r->data.SOA->serial = _dns_parse_uint32(x);
			r->data.SOA->refresh = _dns_parse_uint32(x);
			r->data.SOA->retry = _dns_parse_uint32(x);
			r->data.SOA->expire = _dns_parse_uint32(x);
			r->data.SOA->minimum = _dns_parse_uint32(x);
			break;

		case ns_t_wks:
			if (*remaining < 5)
			{
				dns_free_resource_record(r);
				return NULL;
			}

			*remaining -= 5;
            
            int32_t mapsize = rdlen - 5;
            if (mapsize < 0 || *remaining < mapsize) {
                dns_free_resource_record(r);
                return NULL;
            }
            *remaining -= mapsize;

			size = sizeof(dns_WKS_record_t);
			r->data.WKS = (dns_WKS_record_t *)calloc(1, size);

			r->data.WKS->addr.s_addr = htonl(_dns_parse_uint32(x));
			r->data.WKS->protocol = _dns_parse_uint8(x);
			size = (uint32_t)mapsize;
			r->data.WKS->maplength = size * 8;
			r->data.WKS->map = NULL;
			if (size == 0) break;

			r->data.WKS->map = (uint8_t *)calloc(1, r->data.WKS->maplength);
			mi = 0;
			for (bx = 0; bx < size; bx++)
			{
				byte = _dns_parse_uint8(x);
				for (i = 128; i >= 1; i = i/2)
				{
					if (byte & i) r->data.WKS->map[mi] = 0xff;
					else r->data.WKS->map[mi] = 0;
					mi++;
				}
			}
			break;

		case ns_t_hinfo:
			size = sizeof(dns_HINFO_record_t);
			r->data.HINFO = (dns_HINFO_record_t *)calloc(1, size);

			r->data.HINFO->cpu = _dns_parse_string(p, x, remaining);
			if (r->data.HINFO->cpu == NULL)
			{
				dns_free_resource_record(r);
				return NULL;
			}

			r->data.HINFO->os = _dns_parse_string(p, x, remaining);
			if (r->data.HINFO->os == NULL)
			{
				dns_free_resource_record(r);
				return NULL;
			}

			break;

		case ns_t_minfo:
			size = sizeof(dns_MINFO_record_t);
			r->data.MINFO = (dns_MINFO_record_t *)calloc(1, size);

			r->data.MINFO->rmailbx = _dns_parse_domain_name(p, x, remaining);
			if (r->data.MINFO->rmailbx == NULL)
			{
				dns_free_resource_record(r);
				return NULL;
			}

			r->data.MINFO->emailbx = _dns_parse_domain_name(p, x, remaining);
			if (r->data.MINFO->emailbx == NULL)
			{
				dns_free_resource_record(r);
				return NULL;
			}

			break;

		case ns_t_mx:
			if (*remaining < 2)
			{
				dns_free_resource_record(r);
				return NULL;
			}

			*remaining -= 2;

			size = sizeof(dns_MX_record_t);
			r->data.MX = (dns_MX_record_t *)calloc(1, size);

			r->data.MX->preference = _dns_parse_uint16(x);
			r->data.MX->name = _dns_parse_domain_name(p, x, remaining);
			if (r->data.MX->name == NULL)
			{
				dns_free_resource_record(r);
				return NULL;
			}

			break;

		case ns_t_txt:
			size = sizeof(dns_TXT_record_t);
			r->data.TXT = (dns_TXT_record_t *)malloc(size);
			r->data.TXT->string_count = 0;
			r->data.TXT->strings = NULL;

			while (*x < (eor + rdlen))
			{
				if (r->data.TXT->string_count == 0)
				{
					r->data.TXT->strings = (char **)calloc(1, sizeof(char *));
				}
				else
				{
					r->data.TXT->strings = (char **)realloc(r->data.TXT->strings, (r->data.TXT->string_count + 1) * sizeof(char *));
				}

				r->data.TXT->strings[r->data.TXT->string_count] = _dns_parse_string(p, x, remaining);
				if (r->data.TXT->strings[r->data.TXT->string_count] == NULL)
				{
					dns_free_resource_record(r);
					return NULL;
				}
				r->data.TXT->string_count++;
			}

			break;

		case ns_t_rp:
			size = sizeof(dns_RP_record_t);
			r->data.RP = (dns_RP_record_t *)calloc(1, size);

			r->data.RP->mailbox = _dns_parse_domain_name(p, x, remaining);
			if (r->data.RP->mailbox == NULL)
			{
				dns_free_resource_record(r);
				return NULL;
			}

			r->data.RP->txtdname = _dns_parse_domain_name(p, x, remaining);
			if (r->data.RP->txtdname == NULL)
			{
				dns_free_resource_record(r);
				return NULL;
			}

			break;

		case ns_t_afsdb:
			if (*remaining < 4)
			{
				dns_free_resource_record(r);
				return NULL;
			}

			*remaining -= 4;
			size = sizeof(dns_AFSDB_record_t);
			r->data.AFSDB = (dns_AFSDB_record_t *)calloc(1, size);

			r->data.AFSDB->subtype = _dns_parse_uint32(x);
			r->data.AFSDB->hostname = _dns_parse_domain_name(p, x, remaining);
			if (r->data.AFSDB->hostname == NULL)
			{
				dns_free_resource_record(r);
				return NULL;
			}

			break;

		case ns_t_x25:
			size = sizeof(dns_X25_record_t);
			r->data.X25 = (dns_X25_record_t *)calloc(1, size);

			r->data.X25->psdn_address = _dns_parse_string(p, x, remaining);
			if (r->data.X25->psdn_address == NULL)
			{
				dns_free_resource_record(r);
				return NULL;
			}

			break;

		case ns_t_isdn:
			size = sizeof(dns_ISDN_record_t);
			r->data.ISDN = (dns_ISDN_record_t *)calloc(1, size);

			r->data.ISDN->isdn_address = _dns_parse_string(p, x, remaining);
			if (r->data.ISDN->isdn_address == NULL)
			{
				dns_free_resource_record(r);
				return NULL;
			}

			if (*x < (eor + rdlen))
			{
				r->data.ISDN->subaddress = _dns_parse_string(p, x, remaining);
				if (r->data.ISDN->subaddress == NULL)
				{
					dns_free_resource_record(r);
					return NULL;
				}
			}
			else
			{
				r->data.ISDN->subaddress = NULL;
			}

			break;

		case ns_t_rt:
			if (*remaining < 2)
			{
				dns_free_resource_record(r);
				return NULL;
			}

			*remaining -= 2;

			size = sizeof(dns_RT_record_t);
			r->data.RT = (dns_RT_record_t *)calloc(1, size);

			r->data.RT->preference = _dns_parse_uint16(x);
			r->data.RT->intermediate = _dns_parse_domain_name(p, x, remaining);
			if (r->data.RT->intermediate == NULL)
			{
				dns_free_resource_record(r);
				return NULL;
			}

			break;

		case ns_t_loc:
			if (*remaining < 16)
			{
				dns_free_resource_record(r);
				return NULL;
			}

			*remaining -= 16;

			size = sizeof(dns_LOC_record_t);
			r->data.LOC = (dns_LOC_record_t *)calloc(1, size);

			r->data.LOC->version = _dns_parse_uint8(x);
			r->data.LOC->size = _dns_parse_uint8(x);
			r->data.LOC->horizontal_precision = _dns_parse_uint8(x);
			r->data.LOC->vertical_precision = _dns_parse_uint8(x);
			r->data.LOC->latitude = _dns_parse_uint32(x);
			r->data.LOC->longitude = _dns_parse_uint32(x);
			r->data.LOC->altitude = _dns_parse_uint32(x);
			break;

		case ns_t_srv:
			if (*remaining < 6)
			{
				dns_free_resource_record(r);
				return NULL;
			}

			*remaining -= 6;

			size = sizeof(dns_SRV_record_t);
			r->data.SRV = (dns_SRV_record_t *)calloc(1, size);

			r->data.SRV->priority = _dns_parse_uint16(x);
			r->data.SRV->weight = _dns_parse_uint16(x);
			r->data.SRV->port = _dns_parse_uint16(x);
			r->data.SRV->target = _dns_parse_domain_name(p, x, remaining);
			if (r->data.SRV->target == NULL)
			{
				dns_free_resource_record(r);
				return NULL;
			}

			break;

		case ns_t_null:
		default:
			*remaining -= rdlen;

			size = sizeof(dns_raw_resource_record_t);
			r->data.DNSNULL = (dns_raw_resource_record_t *)calloc(1, size);

			r->data.DNSNULL->length = rdlen;
			r->data.DNSNULL->data = calloc(1, rdlen);
			memmove(r->data.DNSNULL->data, *x, rdlen);
			*x += rdlen;
			break;
	}

    if (rem < *remaining) {
        dns_free_resource_record(r);
        return NULL;
    }
    int32_t bytesread = rem - *remaining;
    int32_t skip = MAX(rdlen, bytesread);
    if (rem < skip) {
        dns_free_resource_record(r);
        return NULL;
    }
    
    *x = eor + skip;
    *remaining = rem - skip;
	return r;
}

static
dns_question_t *
_dns_parse_question_internal(const char *p, char **x, int32_t *remaining)
{
	dns_question_t *q;

	if (x == NULL) return NULL;
	if (*x == NULL) return NULL;
	if (*remaining < 1) return NULL;

	q = (dns_question_t *)calloc(1, sizeof(dns_question_t));

	q->name = _dns_parse_domain_name(p, x, remaining);
	if (q->name == NULL)
	{
		free(q);
		return NULL;
	}

	if (*remaining < 4)
	{
		free(q->name);
		free(q);
		return NULL;
	}

	*remaining = *remaining - 4;

	q->dnstype = _dns_parse_uint16(x);
	q->dnsclass = _dns_parse_uint16(x);

	return q;
}

dns_reply_t *
dns_parse_packet(const char *p, uint32_t len)
{
	dns_reply_t *r;
	dns_header_t *h;
	char *x;
	uint32_t i, size;
    int32_t remaining;

	if (p == NULL) return NULL;
	if (len < NS_HFIXEDSZ) return NULL;

	x = (char *)p;

	r = (dns_reply_t *)calloc(1, sizeof(dns_reply_t));

	r->header = (dns_header_t *)calloc(1, sizeof(dns_header_t));
	h = r->header;

	h->xid = _dns_parse_uint16(&x);
	h->flags = _dns_parse_uint16(&x);
	h->qdcount = _dns_parse_uint16(&x);
	h->ancount = _dns_parse_uint16(&x);
	h->nscount = _dns_parse_uint16(&x);
	h->arcount = _dns_parse_uint16(&x);

	remaining = len - NS_HFIXEDSZ;

	size = sizeof(dns_question_t *);
	r->question = (dns_question_t **)calloc(h->qdcount, size);
	for (i = 0; i < h->qdcount; i++)
	{
		r->question[i] = _dns_parse_question_internal(p, &x, &remaining);
		if (r->question[i] ==NULL)
		{
			h->qdcount = 0;
			if (i > 0) h->qdcount = i - 1;
			h->ancount = 0;
			h->nscount = 0;
			h->arcount = 0;
			dns_free_reply(r);
			return NULL;
		}
	}

	size = sizeof(dns_resource_record_t *);

	r->answer = (dns_resource_record_t **)calloc(h->ancount, size);
	for (i = 0; i < h->ancount; i++)
	{
		r->answer[i] = _dns_parse_resource_record_internal(p, &x, &remaining);
		if (r->answer[i] == NULL)
		{
			h->ancount = 0;
			if (i > 0) h->ancount = i - 1;
			h->nscount = 0;
			h->arcount = 0;
			dns_free_reply(r);
			return NULL;
		}
	}

	r->authority = (dns_resource_record_t **)calloc(h->nscount, size);
	for (i = 0; i < h->nscount; i++)
	{
		r->authority[i] = _dns_parse_resource_record_internal(p, &x, &remaining);
		if (r->authority[i] == NULL)
		{
			h->nscount = 0;
			if (i > 0) h->nscount = i - 1;
			h->arcount = 0;
			dns_free_reply(r);
			return NULL;
		}
	}

	r->additional = (dns_resource_record_t **)calloc(h->arcount, size);
	for (i = 0; i < h->arcount; i++)
	{
		r->additional[i] = _dns_parse_resource_record_internal(p, &x, &remaining);
		if (r->additional[i] == NULL)
		{
			h->arcount = 0;
			if (i > 0) h->arcount = i - 1;
			dns_free_reply(r);
			return NULL;
		}
	}

	return r;
}

void
dns_free_resource_record(dns_resource_record_t *r)
{
	int i;

	free(r->name);

	switch (r->dnstype)
	{
		case ns_t_a:
			if (r->data.A != NULL) free(r->data.A);
			break;

		case ns_t_aaaa:
			if (r->data.AAAA != NULL) free(r->data.AAAA);
			break;

		case ns_t_ns:
		case ns_t_md:
		case ns_t_mf:
		case ns_t_cname:
		case ns_t_mb:
		case ns_t_mg:
		case ns_t_mr:
		case ns_t_ptr:
			if (r->data.CNAME != NULL)
			{
				if (r->data.CNAME->name != NULL) free(r->data.CNAME->name);
				free(r->data.CNAME);
			}
			break;

		case ns_t_soa:
			if (r->data.SOA != NULL)
			{
				if (r->data.SOA->mname != NULL) free(r->data.SOA->mname);
				if (r->data.SOA->rname != NULL) free(r->data.SOA->rname);
				free(r->data.SOA);
			}
			break;

		case ns_t_wks:
			if (r->data.WKS != NULL)
			{
				if (r->data.WKS->map != NULL) free(r->data.WKS->map);
				free(r->data.WKS);
			}
			break;

		case ns_t_hinfo:
			if (r->data.HINFO != NULL)
			{
				if (r->data.HINFO->cpu != NULL) free(r->data.HINFO->cpu);
				if (r->data.HINFO->os != NULL) free(r->data.HINFO->os);
				free(r->data.HINFO);
			}
			break;

		case ns_t_minfo:
			if (r->data.MINFO != NULL)
			{
				if (r->data.MINFO->rmailbx != NULL) free(r->data.MINFO->rmailbx);
				if (r->data.MINFO->emailbx != NULL) free(r->data.MINFO->emailbx);
				free(r->data.MINFO);
			}
			break;

		case ns_t_mx:
			if (r->data.MX != NULL)
			{
				if (r->data.MX->name != NULL) free(r->data.MX->name);
				free(r->data.MX);
			}
			break;


		case ns_t_txt:
			if (r->data.TXT != NULL)
			{
				for (i = 0; i < r->data.TXT->string_count; i++) free(r->data.TXT->strings[i]);
				if (r->data.TXT->strings != NULL) free(r->data.TXT->strings);
				free(r->data.TXT);
			}
			break;

		case ns_t_rp:
			if (r->data.RP != NULL)
			{
				if (r->data.RP->mailbox != NULL) free(r->data.RP->mailbox);
				if (r->data.RP->txtdname != NULL) free(r->data.RP->txtdname);
				free(r->data.RP);
			}
			break;

		case ns_t_afsdb:
			if (r->data.AFSDB != NULL)
			{
				if (r->data.AFSDB->hostname != NULL) free(r->data.AFSDB->hostname);
				free(r->data.AFSDB);
			}
			break;

		case ns_t_x25:
			if (r->data.X25 != NULL)
			{
				if (r->data.X25->psdn_address != NULL) free(r->data.X25->psdn_address);
				free(r->data.X25);
			}
			break;

		case ns_t_isdn:
			if (r->data.ISDN != NULL)
			{
				if (r->data.ISDN->isdn_address != NULL) free(r->data.ISDN->isdn_address);
				if (r->data.ISDN->subaddress != NULL) free(r->data.ISDN->subaddress);
				free(r->data.ISDN);
			}
			break;

		case ns_t_rt:
			if (r->data.RT != NULL)
			{
				if (r->data.RT->intermediate != NULL) free(r->data.RT->intermediate);
				free(r->data.RT);
			}
			break;

		case ns_t_loc:
			if (r->data.LOC != NULL) free(r->data.LOC);
			break;

		case ns_t_srv:
			if (r->data.SRV != NULL)
			{
				if (r->data.SRV->target != NULL) free(r->data.SRV->target);
				free(r->data.SRV);
			}
			break;

		case ns_t_invalid:
			break;

		case ns_t_null:
		default:
			if (r->data.DNSNULL != NULL)
			{
				if (r->data.DNSNULL->data != NULL) free(r->data.DNSNULL->data);
				free(r->data.DNSNULL);
			}
			break;
	}

	free(r);
}

void
dns_free_reply(dns_reply_t *r)
{
	uint32_t i;

	if (r == NULL) return;
	if (r->header != NULL)
	{
		for (i = 0; i < r->header->qdcount; i++)
		{
			free(r->question[i]->name);
			free(r->question[i]);
		}

		for (i = 0; i < r->header->ancount; i++) dns_free_resource_record(r->answer[i]);
		for (i = 0; i < r->header->nscount; i++) dns_free_resource_record(r->authority[i]);
		for (i = 0; i < r->header->arcount; i++) dns_free_resource_record(r->additional[i]);

		free(r->header);
	}

	if (r->question != NULL) free(r->question);
	if (r->answer != NULL) free(r->answer);
	if (r->authority != NULL) free(r->authority);
	if (r->additional != NULL) free(r->additional);

	if (r->server != NULL) free(r->server);

	free(r);
}
