/*
 * Copyright (c) 2003-2006 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 * 
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */


#ifndef __DNS_UTIL_H__
#define __DNS_UTIL_H__

#ifdef __cplusplus
extern "C"{
#endif

#include <arpa/inet.h>

/*
 * DNS query / reply header
 */
typedef struct {
	uint16_t xid;
	uint16_t flags;
	uint16_t qdcount;
	uint16_t ancount;
	uint16_t nscount;
	uint16_t arcount;
} dns_header_t;

/*
 * DNS query
 */
typedef struct
{
	char *name;
	uint16_t dnstype;
	uint16_t dnsclass;
} dns_question_t;

/*
 * Resource Record types
 * dns_parse_packet() creates resourse records of these types.
 */
typedef struct
{
	uint16_t length;
	char *data;
} dns_raw_resource_record_t;

typedef struct
{
	struct in_addr addr;
} dns_address_record_t;

typedef struct
{
	struct in6_addr addr;
} dns_in6_address_record_t;

typedef struct
{
	char *name;
} dns_domain_name_record_t;

typedef struct
{
	char *mname;
	char *rname;
	uint32_t serial;
	uint32_t refresh;
	uint32_t retry;
	uint32_t expire;
	uint32_t minimum;
} dns_SOA_record_t;

typedef struct
{
	char *cpu;
	char *os;
} dns_HINFO_record_t;

typedef struct
{
	char *rmailbx;
	char *emailbx;
} dns_MINFO_record_t;

typedef struct
{
	uint16_t preference;
	char *name;
} dns_MX_record_t;

typedef struct
{
	uint32_t string_count;
	char **strings;
} dns_TXT_record_t;

typedef struct
{
	struct in_addr addr;
	uint8_t protocol;
	uint32_t maplength;
	uint8_t *map;
} dns_WKS_record_t;

typedef struct
{
	char *mailbox;
	char *txtdname;
} dns_RP_record_t;

typedef struct
{
	uint32_t subtype;
	char *hostname;
} dns_AFSDB_record_t;

typedef struct
{
	char *psdn_address;
} dns_X25_record_t;

typedef struct
{
	char *isdn_address;
	char *subaddress;
} dns_ISDN_record_t;

typedef struct
{
	uint16_t preference;
	char * intermediate;
} dns_RT_record_t;

typedef struct
{
	uint8_t version;
	uint8_t size;
	uint8_t horizontal_precision;
	uint8_t vertical_precision;
	uint32_t latitude;
	uint32_t longitude;
	uint32_t altitude;
} dns_LOC_record_t;

typedef struct
{
	uint16_t priority;
	uint16_t weight;
	uint16_t port;
	char *target;
} dns_SRV_record_t;

/*
 * DNS Resource Record
 *
 * Data contained in unsupported or obsolete Resource Record types
 * may be accessed via DNSNULL as a dns_raw_resource_record_t. 
 */
typedef struct
{
	char *name;
	uint16_t dnstype;
	uint16_t dnsclass;
	uint32_t ttl;
	union
	{
		dns_address_record_t *A;
		dns_domain_name_record_t *NS;
		dns_domain_name_record_t *MD;		/* Obsolete */
		dns_domain_name_record_t *MF;		/* Obsolete */
		dns_domain_name_record_t *CNAME;
		dns_SOA_record_t *SOA;
		dns_domain_name_record_t *MB;
		dns_domain_name_record_t *MG;
		dns_domain_name_record_t *MR;
		dns_raw_resource_record_t *DNSNULL;
		dns_WKS_record_t *WKS;
		dns_domain_name_record_t *PTR;
		dns_HINFO_record_t *HINFO;
		dns_MINFO_record_t *MINFO;
		dns_MX_record_t *MX;
		dns_TXT_record_t *TXT;
		dns_RP_record_t *RP;
		dns_AFSDB_record_t *AFSDB;
		dns_X25_record_t *X25;
		dns_ISDN_record_t *ISDN;
		dns_RT_record_t *RT;
		dns_in6_address_record_t *AAAA;
		dns_LOC_record_t *LOC;
		dns_SRV_record_t *SRV;
	} data;
} dns_resource_record_t;

/*
 * A parsed DNS record.  Returned by dns_parse_packet() and dns_lookup().
 * The contents may be printed using dns_print_reply().
 */
typedef struct
{
	uint32_t status;
	struct sockaddr *server;
	dns_header_t *header;
	dns_question_t **question;
	dns_resource_record_t **answer;
	dns_resource_record_t **authority;
	dns_resource_record_t **additional;
} dns_reply_t;

/*
 * Parse a reply packet into a reply structure.
 */
dns_reply_t *dns_parse_packet(const char *buf, uint32_t len);

/*
 * Free a reply structure.
 */
void dns_free_reply(dns_reply_t *r);

#ifdef __cplusplus
}
#endif

#endif /* __DNS_UTIL_H__ */
